"""
Functions for writing to pvsite db
"""

import datetime as dt
import logging
import uuid

import dateutil.parser as dp
import numpy as np
import pandas as pd
import sqlalchemy.orm as sa_orm
from pvsite_datamodel.sqlmodels import GenerationSQL

# Defines the length of time over which a forecast is valid
from pvsite_datamodel.write.datetime_intervals import get_or_else_create_datetime_interval
from pvsite_datamodel.write.upsert import upsert
from pvsite_datamodel.write.utils import WrittenRow
from pvsite_datamodel.read import get_site_from_uuid



def insert_generation_values(
    session: sa_orm.Session, generation_values_df: pd.DataFrame,
) -> list[WrittenRow]:
    """
    Inserts a dataframe of forecast values into the database.

    :param session: sqlalchemy session for interacting with the database
    :param generation_values_df: pandas dataframe with columns
    ["start_datetime_utc", "power_kw", "pv_uuid"]
    :return list[WrittenRow]: list of added rows to DB
    """

    # Track rows added to DB
    written_rows: list[WrittenRow] = []

    # Loop over all the unique sites that have got forecast values
    site_uuids: np.ndarray = generation_values_df["site_uuid"].unique()
    generation_sqls = []
    for site_uuid in site_uuids:

        # Check whether the site id exits in the table, otherwise return an error
        get_site_from_uuid(session=session, site_uuid=site_uuid)

        # Get all dataframe forecast value entries for current site_uuid
        df_site: pd.DataFrame = generation_values_df.loc[generation_values_df["site_uuid"] == site_uuid]

        # Filter the forecasted values by target_time
        start_datetimes: np.ndarray = df_site["start_datetime_utc"].unique()

        # Print a warning if there are duplicate target_times for this site's forecast
        if len(start_datetimes) != len(df_site):
            logging.warning(
                f"duplicate target datetimes "
                f"for site {site_uuid}"
            )

        # For each target time:
        for start_datetime in start_datetimes:

            datetime_interval, newly_added_rows = get_or_else_create_datetime_interval(
                session=session, start_time=pd.to_datetime(start_datetime)
            )
            written_rows.extend(newly_added_rows)

            # For each entry with this target time:
            df_target_entries: pd.DataFrame = df_site.loc[
                df_site["start_datetime_utc"] == start_datetimes
            ]

            # Create a GenerationSQL object for each generation, and surface as dict
            generation = GenerationSQL(
                site_uuid=site_uuid,
                generation_uuid=uuid.uuid4(),
                power_kw=df_target_entries.iloc[0].power_kw,
                datetime_interval_uuid=datetime_interval.datetime_interval_uuid,
            ).__dict__
            generation_sqls.append(generation)

        # Save it to the db
        newly_added_rows = upsert(session, GenerationSQL, generation_sqls)
        written_rows.extend(newly_added_rows)

    return written_rows





import datetime as dt
import typing

from pvsite_datamodel import sqlmodels


class WrittenRow(typing.NamedTuple):
    """
    Defines a write to the Database
    """

    table: sqlmodels.Base
    pk_value: str


FORECAST_TIMESPAN = dt.timedelta(minutes=5)


"""
SQLAlchemy definition of the pvsite database schema
"""

import uuid
from datetime import datetime

import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import FLOAT, INTEGER, REAL, TIMESTAMP, UUID
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy.schema import UniqueConstraint
from sqlalchemy import Column, DateTime

Base = declarative_base()


class CreatedMixin:
    """Mixin to add created datetime to model"""

    created_utc = Column(DateTime(timezone=True), default=lambda: datetime.utcnow())


class SiteSQL(Base, CreatedMixin):
    """
    Class representing the sites table.

    Each site row specifies a single panel or cluster of panels
    found on a residential house or commercial building. Their
    data is provided by a client.

    *Approximate size: *
    4 clients * ~1000 sites each = ~4000 rows
    """

    __tablename__ = "sites"

    site_uuid = sa.Column(UUID(as_uuid=True), default=uuid.uuid4, primary_key=True)
    client_uuid = sa.Column(
        UUID(as_uuid=True), sa.ForeignKey("clients.client_uuid"), nullable=False, default=uuid.uuid4
    )
    client_site_id = sa.Column(INTEGER)
    client_site_name = sa.Column(sa.String(255))
    region = sa.Column(sa.String(255))
    dno = sa.Column(sa.String(255))
    gsp = sa.Column(sa.String(255))
    orientation = sa.Column(REAL)
    tilt = sa.Column(REAL)
    latitude = sa.Column(FLOAT, nullable=False)
    longitude = sa.Column(FLOAT, nullable=False)
    capacity_kw = sa.Column(REAL, nullable=False)
    updated_utc = sa.Column(TIMESTAMP, nullable=False, default=lambda: datetime.utcnow())
    ml_id = sa.Column(sa.Integer, autoincrement=True, nullable=False, unique=True)

    __table_args__ = (UniqueConstraint("client_site_id", client_uuid, name="idx_client"),)

    forecasts = relationship("ForecastSQL")
    latest_forecast_values = relationship("LatestForecastValueSQL")
    generation = relationship("GenerationSQL")


class GenerationSQL(Base, CreatedMixin):
    """
    Class representing the generation table.

    Each generation row specifies a generated power output over a
    given time range for a site.

    *Approximate size: *
    Generation populated every 5 minutes per site * 4000 sites = ~1,125,000 rows per day
    """

    __tablename__ = "generation"

    generation_uuid = sa.Column(UUID(as_uuid=True), default=uuid.uuid4, primary_key=True)
    site_uuid = sa.Column(
        UUID(as_uuid=True), sa.ForeignKey("sites.site_uuid"), nullable=False, default=uuid.uuid4
    )
    power_kw = sa.Column(REAL, nullable=False)
    datetime_interval_uuid = sa.Column(
        UUID(as_uuid=True),
        sa.ForeignKey("datetime_intervals.datetime_interval_uuid"),
        nullable=False,
        default=uuid.uuid4,
    )


class ForecastSQL(Base, CreatedMixin):
    """
    Class representing the forecasts table.

    Each forecast row refers to a sequence of predicted solar generation values
    over a set of target times for a site.

    *Approximate size: *
    One forecast per site every 5 minutes = ~1,125,000 rows per day
    """

    __tablename__ = "forecasts"

    forecast_uuid = sa.Column(UUID(as_uuid=True), default=uuid.uuid4, primary_key=True)
    site_uuid = sa.Column(
        UUID(as_uuid=True), sa.ForeignKey("sites.site_uuid"), nullable=False, default=uuid.uuid4
    )

    forecast_version = sa.Column(sa.String(32), nullable=False)

    # one (forecasts) to many (forecast_values)
    forecast_values = relationship("ForecastValueSQL")


class ForecastValueSQL(Base, CreatedMixin):
    """
    Class representing the forecast_values table.

    Each forecast_value row is a prediction for the power output
    of a site over a target datetime interval. Many predictions
    are made for each site at each target interval.

    *Approximate size: *
    One forecast value every 5 minutes per site per forecast.
    Each forecast's prediction sequence covers 24 hours of target
    intervals = ~324,000,000 rows per day
    """

    __tablename__ = "forecast_values"

    forecast_value_uuid = sa.Column(UUID(as_uuid=True), default=uuid.uuid4, primary_key=True)
    datetime_interval_uuid = sa.Column(
        UUID(as_uuid=True),
        sa.ForeignKey("datetime_intervals.datetime_interval_uuid"),
        nullable=False,
        default=uuid.uuid4,
    )
    forecast_generation_kw = sa.Column(REAL, nullable=False)

    forecast_uuid = sa.Column(
        UUID(as_uuid=True),
        sa.ForeignKey("forecasts.forecast_uuid"),
        nullable=False,
        default=uuid.uuid4,
    )


class LatestForecastValueSQL(Base, CreatedMixin):
    """
    Class representing the latest_forecast_values table.

    Each latest_forecast_value row is a prediction for the power output
    of a site over a target datetime interval. Only the most recent
    prediction for each target time interval is stored in this table
    per site.

    *Approximate size: *
    One forecast value every 5 minutes per site per forecast
    sequence = ~1,125,000 rows per day
    """

    __tablename__ = "latest_forecast_values"

    latest_forecast_value_uuid = sa.Column(UUID(as_uuid=True), default=uuid.uuid4, primary_key=True)
    datetime_interval_uuid = sa.Column(
        UUID(as_uuid=True),
        sa.ForeignKey("datetime_intervals.datetime_interval_uuid"),
        nullable=False,
        default=uuid.uuid4,
    )
    forecast_generation_kw = sa.Column(REAL, nullable=False)

    forecast_uuid = sa.Column(UUID(as_uuid=True), default=uuid.uuid4, nullable=False)
    site_uuid = sa.Column(
        UUID(as_uuid=True), sa.ForeignKey("sites.site_uuid"), nullable=False, default=uuid.uuid4
    )
    forecast_version = sa.Column(sa.String(32), nullable=False)

    latest_forecast = relationship("SiteSQL", back_populates="latest_forecast_values")


class ClientSQL(Base, CreatedMixin):
    """
    Class representing the clients table.

    Each client row defines a provider of site data

    *Approximate size: *
    One row per client = ~4 rows
    """

    __tablename__ = "clients"

    client_uuid = sa.Column(UUID(as_uuid=True), default=uuid.uuid4, primary_key=True)
    client_name = sa.Column(sa.String(255), nullable=False)

    sites = relationship("SiteSQL")


class DatetimeIntervalSQL(Base, CreatedMixin):
    """
    Class representing the datetime_intervals table.

    Each datetime_interval row defines a timespan between a start and end time

    *Approximate size: *
    One interval every 5 minutes per day = ~288 rows per day
    """

    __tablename__ = "datetime_intervals"

    datetime_interval_uuid = sa.Column(UUID(as_uuid=True), default=uuid.uuid4, primary_key=True)
    start_utc = sa.Column(TIMESTAMP, nullable=False)
    end_utc = sa.Column(TIMESTAMP, nullable=False)

    generation = relationship("GenerationSQL")
    forecast_values = relationship("ForecastValueSQL")
    latest_forecast_values = relationship("LatestForecastValueSQL")


class StatusSQL(Base, CreatedMixin):
    """
    Class representing the status table:

    Each status row defines a message reporting on the status of the
    services within the nowcasting domain

    *Approximate size: *
    ~1 row per day
    """

    __tablename__ = "status"

    status_uuid = sa.Column(UUID(as_uuid=True), default=uuid.uuid4, primary_key=True)
    status = sa.Column(sa.String(255))
    message = sa.Column(sa.String(255))



"""
Functions for reading to pvsite db
"""

import logging

from sqlalchemy.orm import Session
from pvsite_datamodel.sqlmodels import ClientSQL, SiteSQL

logger = logging.getLogger(__name__)


def get_site_from_uuid(session: Session, site_uuid: str):
    """
    Get site object from uuid.

    Raise error if site does not exists

    :param session: database sessions
    :param site_uuid: the site uuid
    :return: the site object
    """
    query = session.query(SiteSQL)
    query = query.filter(SiteSQL.site_uuid == site_uuid)
    existing_site: SiteSQL = query.first()
    if existing_site is None:
        raise KeyError(f"Site uuid {site_uuid} not found in sites table")

    return existing_site


def get_site(
    session: Session, client_name: str, client_id: str
) -> SiteSQL:
    """
    Get site from client name and client id

    :param session: database sessions
    :param client_name: client name
    :param client_id: client id
    :return: site object
    """

    logger.debug(f'Getting site for {client_name=} and {client_id=}')

    # start main query
    query = session.query(SiteSQL)
    query = query.join(ClientSQL)

    # select the correct client id
    query = query.filter(SiteSQL.client_site_id == client_id)

    # filter on client_name
    query = query.filter(ClientSQL.client_name == client_name)

    # get all results
    site = query.first()

    if site is None:
        raise Exception(f"Could not find site with {client_id=} and {client_name=}")

    return site


def get_all_site(
    session: Session,
) -> SiteSQL:
    """
    Get all sites

    :param session: database sessions
    :return: site object
    """

    logger.debug(f'Getting all sites')

    # start main query
    query = session.query(SiteSQL)

    # get all results
    sites = query.all()

    logger.debug(f'Found {len(sites)} sites')

    return sites

